

// ******* LUA demo ******* //
// version 1
// check manual at http://www.lua.org/manual/5.1/manual.html


#include <iostream>
#include "lua_src/lua.hpp"
#include <exception>
#include <typeinfo>
#include <string>
using namespace std;

// Allocation function for lua (required to create it's state machine object)
// http://www.lua.org/manual/5.1/manual.html#lua_Alloc
void *luaAlloc (void *customData, void *dynMemory, size_t oldSize, size_t newSize);

// Panic function, called when there is an error
// The built-in panic function exits the application. You must create your own which throws
// an exception
// http://www.lua.org/manual/5.1/manual.html#lua_atpanic
int noPanic( lua_State *L );

// A simple C function to be exported into lua. It accepts no arguments and returns no value.
// http://www.lua.org/manual/5.1/manual.html#lua_CFunction
static int helloLua(lua_State *L);

// Extended C function for lua. It accepts two arguments: number and string.
// Lua function can return multiple values
// It returns them in reverse order: first string, then number
static int complexFunction(lua_State *L);

// object to be referenced from lua
struct Entry{
	const char* name;
	double value;

	Entry(const char* name, double value){
		this->name=name;
		this->value=value;
	}
};

// A C function to be called from lua with the Entry object as argument
static int readEntry(lua_State *L);

int main(){
	// Create the lua state machine object. It keeps the context of the scripts until it's deleted.
	// http://www.lua.org/manual/5.1/manual.html#luaL_newstate
	lua_State *L = lua_newstate(luaAlloc, NULL);
	// set panic function
	lua_atpanic(L, noPanic);
	// open standart libs
	luaL_openlibs(L);

	try{
		// Register the helloLua function
		lua_pushcfunction(L,helloLua);
		// in lua, functions are variables
		// create it as global variable
		lua_setglobal(L,"helloLua");

		// Register the complexFunction
		lua_pushcfunction(L,complexFunction);
		lua_setglobal(L,"complexFunction");

		// create Entry object and export it into lua as light userdata
		Entry* lightEntry = new Entry("This year", 2009.0);
		lua_pushlightuserdata(L, static_cast<void*>(lightEntry));
		lua_setglobal(L, "lightEntry");

		// crreate another Entry, as full userdata
		void* fullEntryV = lua_newuserdata(L, sizeof(Entry));
		lua_setglobal(L, "fullEntry");
		if(fullEntryV==NULL){
			throw string("Allocation of full userdata failed");
		}
		Entry* fullEntry = static_cast<Entry*>(fullEntryV);
		fullEntry->name = "Last year";
		fullEntry->value = 2008.0;

		// export readEntry function
		lua_pushcfunction(L, readEntry);
		lua_setglobal(L, "readEntry");

		// parse the script file
		int res = luaL_dofile(L,"script.lua");
		if(res!=0){
			// on error, get the error message and throw
			throw string( lua_tostring(L, lua_gettop(L)) );
		}

		// read the global variable textFromWithin
		lua_getglobal(L,"textFromWithin");// push the variable to the stack
		const char* textFromWithin = lua_tostring(L,lua_gettop(L)); // read it from the stack
		cout<<"Lua global variable 'textFromWithin' : "<<textFromWithin<<endl;

		// Read value from the lua table
		// ob.sub.substr string
		lua_getglobal(L, "ob"); // push the global table to stack
		int obIndex = lua_gettop(L); // get the global table's stack index
		lua_getfield(L, obIndex, "sub"); // push the subtable on the stack
		int subIndex = lua_gettop(L); // get it's index
		lua_getfield(L, subIndex, "subnum");
		// the subnum attribute
		double subnum = static_cast<double>(luaL_checknumber(L, lua_gettop(L)));
		cout<<"Value retrieved from attribute 'ob.sub.subnum':"<<subnum<<endl;

		// call the helloCPlusPlus lua function
		// all reference at http://www.lua.org/manual/5.1/manual.html#lua_call
		// void lua_call (lua_State *L, int nargs, int nresults);
		lua_settop(L,0);
		lua_getglobal(L,"bigger");
		lua_pushnumber(L, static_cast<lua_Number>(5.0));
		lua_pushnumber(L, static_cast<lua_Number>(15.0));
		lua_call(L, 2, 1);
		// print the result
		cout<<"bigger() returned:"<<static_cast<double>(luaL_checknumber(L, lua_gettop(L)))<<endl;

		// fullEntry gets deleted by garbage collection
		delete lightEntry;

	}
	catch(exception ex){
		cerr<<"Exception: "<<ex.what()<<endl;
	}
	catch(string sx){
		cerr<<"Exception: "<<sx<<endl;
	}
}

// This implementation is copied from the reference manual. Works OK
void *luaAlloc (void *customData, void *dynMemory, size_t oldSize, size_t newSize) {
	if( newSize == 0 ){
		free(dynMemory);
		return NULL;
	}
	else{
		return realloc( dynMemory, newSize );
	}
}
// Panic function throwing std::exception
int noPanic( lua_State *L ){
	std::string msg = "Lua panic: ";
	if( lua_isstring( L, lua_gettop(L) ) ){
		const char *cstr = lua_tostring( L, lua_gettop(L) );
		msg+=cstr;
	}
	else{
		msg+="No message available";
	}
	throw msg;
}

static int helloLua(lua_State *L){
	cout<<"Hello Lua !"<<endl;
	// Lua function must allways return the number of return values
	return 0;
}

static int complexFunction(lua_State *L){
	// Lua state has a virtual stack.
	// C functions for exporting into lua must manipulate the stack to get arguments and return values.
	// Lua functions can return multiple values

	// the main way to read values from lua stack is through their stack index
	// lua_gettop returns the index of the top element.
	// Lua indices always start with 1

	// The number of arguments passed to function is always variable.
	// Get the argument count
	int argumentCount = lua_gettop(L);

	// Get the arguments (in reverse order, the first element was pushed first)

	// the string
	// Uses a different style function to get it
	// this function creates nice error message on invalid value
	const char* stringArgument = luaL_checkstring(L,2);

	// pop the string from the stack
	lua_pop(L,1);

	// Check the argument type, to demonstrate lua error
	// the number
	// http://www.lua.org/manual/5.1/manual.html#lua_gettop
	if( !lua_isnumber(L, lua_gettop(L)) ){
		// push error message
		lua_pushstring(L,"complexFunction: The first argument is not a number");
		// Cause lua panic
		lua_error(L);
	}
	double numericArgument = static_cast<double>( lua_tonumber(L, lua_gettop(L)) );

	// print message
	cout<<"complexFunction: Number: "<<numericArgument<<" String:"<<stringArgument<<endl;

	// Clear the stack and push return values
	lua_settop(L,0);
	lua_pushstring(L, stringArgument);
	lua_pushnumber(L, numericArgument);
	// number of return values
	return 2;
}

static int readEntry(lua_State *L){
	if(lua_isuserdata(L,lua_gettop(L))){
		void* udata = lua_touserdata(L,lua_gettop(L));
		if(udata == NULL){
			throw string("readEntry(): argument is not userdata");
		}
		Entry* e = static_cast<Entry*>(udata);
		cout<<"readEntry(): "<<e->name<<" = "<<e->value<<endl;
		return 0;
	}
	else{
		throw string("readEntry(): argument is not userdata");
	}
}
